@@ -1,3 +1,5 @@
+# mypy: ignore-errors
+
 import datetime
 import io
 import itertools
@@ -15,6 +17,7 @@ from torch.onnx.symbolic_helper import _default_onnx_opset_version
 from torch.onnx.utils import \
     _export as torch_export, _model_to_graph as torch_model_to_graph
 
+from pytorch_pfn_extras.onnx import _as_output as as_output
 from pytorch_pfn_extras.onnx.annotate import init_annotate
 from pytorch_pfn_extras.onnx.strip_large_tensor import \
     LARGE_TENSOR_DATA_THRESHOLD
@@ -114,12 +117,15 @@ def _export(
     if opset_ver is None:
         opset_ver = _default_onnx_opset_version
     strip_doc_string = kwargs.pop('strip_doc_string', True)
-    with init_annotate(model, opset_ver) as ann:
+    with init_annotate(model, opset_ver) as ann, \
+            as_output.trace(model) as (model, outputs):
         outs = _export_util(
             model, args, bytesio, strip_doc_string=False, **kwargs)
         onnx_graph = onnx.load(io.BytesIO(bytesio.getvalue()))
         onnx_graph = ann.set_annotate(onnx_graph)
         onnx_graph = ann.reorg_anchor(onnx_graph)
+        outputs.add_outputs_to_model(onnx_graph)
+
     if strip_doc_string:
         for node in onnx_graph.graph.node:
             node.doc_string = b''
@@ -198,6 +204,9 @@ def export_testcase(
 
     """
 
+    if user_meta is None:
+        user_meta = {}
+
     os.makedirs(out_dir, exist_ok=True)
     input_names = kwargs.pop(
         'input_names',
