@@ -1,28 +1,52 @@
+# mypy: ignore-errors
+
 import atexit
 from contextlib import contextmanager
 import os
 import time
-from typing import Tuple
+from typing import Callable, Dict, Generator, Optional, Tuple
 import threading
 import queue
 import multiprocessing as mp
 import torch
 import weakref
+
 from pytorch_pfn_extras.reporting import DictSummary
 
 
 Events = Tuple[torch.cuda.Event, torch.cuda.Event]
 
 
+class _ReportNotification:
+    def __init__(self, summary, tag, use_cuda, begin_event, begin):
+        self._is_completed = True
+        self._summary = summary
+        self._tag = tag
+        self._use_cuda = use_cuda
+        self._begin_event = begin_event
+        self._begin = begin
+
+    def defer(self):
+        self._is_completed = False
+
+    def complete(self):
+        self._summary.complete_report(
+            self._tag, self._use_cuda, self._begin_event, self._begin)
+
+
 class _CPUWorker:
-    def __init__(self, add, max_queue_size: int):
+    def __init__(
+            self,
+            add: Callable[[str, float], None],
+            max_queue_size: int,
+    ) -> None:
         self._add = add
         self._max_queue_size = max_queue_size
         self._initialized = False
-        self._queue = None
-        self._thread = None
+        self._queue: Optional[mp.JoinableQueue[Optional[Tuple[str, float]]]] = None
+        self._thread: Optional[threading.Thread] = None
 
-    def initialize(self):
+    def initialize(self) -> None:
         if self._initialized:
             return
         self._queue = mp.JoinableQueue(self._max_queue_size)
@@ -30,9 +54,11 @@ class _CPUWorker:
         self._thread.start()
         self._initialized = True
 
-    def finalize(self):
+    def finalize(self) -> None:
         if not self._initialized:
             return
+        assert self._queue is not None
+        assert self._thread is not None
         self._queue.put(None)
         self._queue.join()
         self._queue.close()
@@ -40,15 +66,16 @@ class _CPUWorker:
         self._thread.join()
         self._initialized = False
 
-    def synchronize(self):
-        assert self._initialized
+    def synchronize(self) -> None:
+        assert self._queue is not None
         self._queue.join()
 
-    def put(self, name, value):
-        assert self._initialized
+    def put(self, name: str, value: float) -> None:
+        assert self._queue is not None
         self._queue.put((name, value))
 
-    def _worker(self):
+    def _worker(self) -> None:
+        assert self._queue is not None
         while True:
             v = self._queue.get()
             if v is None:
@@ -59,17 +86,24 @@ class _CPUWorker:
             self._queue.task_done()
 
 
+_QueueElem = Tuple[str, Tuple[torch.cuda.Event, torch.cuda.Event]]
+
+
 class _CUDAWorker:
-    def __init__(self, add, max_queue_size: int):
+    def __init__(
+            self,
+            add: Callable[[str, float], None],
+            max_queue_size: int,
+    ) -> None:
         self._add = add
         self._max_queue_size = max_queue_size
         self._initialized = False
-        self._thread = None
-        self._queue = None
+        self._thread: Optional[threading.Thread] = None
+        self._queue: Optional['queue.Queue[Optional[_QueueElem]]'] = None
         self._event_lock = threading.Lock()
-        self._events = None
+        self._events: Optional['queue.Queue[torch.cuda.Event]'] = None
 
-    def initialize(self):
+    def initialize(self) -> None:
         if self._initialized:
             return
         self._queue = queue.Queue(self._max_queue_size)
@@ -78,31 +112,39 @@ class _CUDAWorker:
         self._thread.start()
         self._initialized = True
 
-    def finalize(self):
+    def finalize(self) -> None:
         if not self._initialized:
             return
+        assert self._queue is not None
+        assert self._thread is not None
         self._queue.put(None)
         self._queue.join()
         self._thread.join()
         self._initialized = False
 
-    def synchronize(self):
-        assert self._initialized
+    def synchronize(self) -> None:
+        assert self._queue is not None
         self._queue.join()
 
-    def put(self, name, events):
-        assert self._initialized
+    def put(
+            self,
+            name: str,
+            events: Tuple[torch.cuda.Event, torch.cuda.Event],
+    ) -> None:
+        assert self._queue is not None
         self._queue.put((name, events))
 
-    def _worker(self):
+    def _worker(self) -> None:
+        assert self._queue is not None
+        assert self._events is not None
         while True:
             v = self._queue.get()
             if v is None:
                 self._queue.task_done()
                 break
             name, (begin, end) = v
-            end.synchronize()
-            t_ms = begin.elapsed_time(end)
+            end.synchronize()  # type: ignore[no-untyped-call]
+            t_ms = begin.elapsed_time(end)  # type: ignore[no-untyped-call]
             self._add(name, t_ms / 1000)
             with self._event_lock:
                 self._events.put(begin)
@@ -111,17 +153,20 @@ class _CUDAWorker:
 
     def get_cuda_event(self) -> torch.cuda.Event:
         assert self._initialized
+        assert self._events is not None
         with self._event_lock:
             if self._events.empty():
-                self._events.put(torch.cuda.Event(enable_timing=True))
+                event = torch.cuda.Event(  # type: ignore[no-untyped-call]
+                    enable_timing=True)
+                self._events.put(event)
             return self._events.get()
 
 
 class _Finalizer:
-    def __init__(self, ts):
+    def __init__(self, ts: 'TimeSummary') -> None:
         self._ts = weakref.ref(ts)
 
-    def __call__(self):
+    def __call__(self) -> None:
         ts = self._ts()
         if ts:
             ts.finalize()
@@ -140,16 +185,15 @@ class TimeSummary:
             when the instance is created.
     """
 
-    def __init__(self, *, max_queue_size: int = 1000, auto_init: bool = True):
+    def __init__(self, *, max_queue_size: int = 1000, auto_init: bool = True) -> None:
         self._summary_lock = threading.Lock()
         self._summary = DictSummary()
-        self._additional_stats = {}
+        self._additional_stats: Dict[str, float] = {}
 
         self._cpu_worker = _CPUWorker(self._add_from_worker, max_queue_size)
+        self._cuda_worker: Optional[_CUDAWorker] = None
         if torch.cuda.is_available():
             self._cuda_worker = _CUDAWorker(self._add_from_worker, max_queue_size)
-        else:
-            self._cuda_worker = None
 
         self._initialized = False
         self._master_pid = os.getpid()
@@ -157,7 +201,7 @@ class TimeSummary:
             self.initialize()
         atexit.register(_Finalizer(self))
 
-    def __del__(self):
+    def __del__(self) -> None:
         self.finalize()
 
     def initialize(self) -> None:
@@ -180,7 +224,7 @@ class TimeSummary:
             self._cuda_worker.initialize()
         self._initialized = True
 
-    def finalize(self):
+    def finalize(self) -> None:
         if not self._initialized:
             return
         self._cpu_worker.finalize()
@@ -188,13 +232,13 @@ class TimeSummary:
             self._cuda_worker.finalize()
         self._initialized = False
 
-    def synchronize(self):
+    def synchronize(self) -> None:
         self.initialize()
         self._cpu_worker.synchronize()
         if self._cuda_worker is not None:
             self._cuda_worker.synchronize()
 
-    def _add_from_worker(self, name, value):
+    def _add_from_worker(self, name: str, value: float) -> None:
         assert self._initialized
         with self._summary_lock:
             self._summary.add({name: value})
@@ -203,11 +247,14 @@ class TimeSummary:
             max_value = self._additional_stats.get(f"{name}.max", value)
             self._additional_stats[f"{name}.max"] = max(value, max_value)
 
-    def add(self, name, value):
+    def add(self, name: str, value: float) -> None:
         self._add_from_worker(name, value)
 
     @contextmanager
-    def summary(self, clear: bool = False):
+    def summary(
+            self,
+            clear: bool = False,
+    ) -> Generator[Tuple[DictSummary, Dict[str, float]], None, None]:
         self.initialize()
         try:
             with self._summary_lock:
@@ -217,8 +264,17 @@ class TimeSummary:
                 self._summary = DictSummary()
                 self._additional_stats = {}
 
+    def complete_report(self, tag, use_cuda, begin_event, begin):
+        end = time.time()
+        self._cpu_worker._queue.put((tag, end - begin))
+        if use_cuda:
+            end_event = self._cuda_worker.get_cuda_event()
+            end_event.record()
+            self._cuda_worker._queue.put(
+                (f"{tag}.cuda", (begin_event, end_event)))
+
     @contextmanager
-    def report(self, tag: str, use_cuda: bool = False) -> None:
+    def report(self, tag: str, use_cuda: bool = False) -> Generator[None, None, None]:
         """Context manager to automatically report execution times.
 
         The start and completion times are obtained automatically,
@@ -230,19 +286,19 @@ class TimeSummary:
             use_cuda (bool): Indicates if GPU time should also be profiled.
         """
         self.initialize()
+        begin_event = None
         if use_cuda:
+            assert self._cuda_worker is not None
             begin_event = self._cuda_worker.get_cuda_event()
-            begin_event.record()
+            begin_event.record()  # type: ignore[no-untyped-call]
         try:
             begin = time.time()
-            yield
+            notification = _ReportNotification(
+                self, tag, use_cuda, begin_event, begin)
+            yield notification
         finally:
-            end = time.time()
-            self._cpu_worker.put(tag, end - begin)
-            if use_cuda:
-                end_event = self._cuda_worker.get_cuda_event()
-                end_event.record()
-                self._cuda_worker.put(f"{tag}.cuda", (begin_event, end_event))
+            if notification._is_completed:
+                self.complete_report(tag, use_cuda, begin_event, begin)
 
 
 time_summary = TimeSummary(auto_init=False)
