@@ -1,4 +1,5 @@
 from pytorch_pfn_extras.training.extension import Extension  # NOQA
+from pytorch_pfn_extras.training.extension import ExtensionEntry  # NOQA
 from pytorch_pfn_extras.training.extension import make_extension  # NOQA
 from pytorch_pfn_extras.training.extension import PRIORITY_EDITOR  # NOQA
 from pytorch_pfn_extras.training.extension import PRIORITY_READER  # NOQA
@@ -6,3 +7,4 @@ from pytorch_pfn_extras.training.extension import PRIORITY_WRITER  # NOQA
 from pytorch_pfn_extras.training import extensions  # NOQA
 from pytorch_pfn_extras.training.manager import ExtensionsManager  # NOQA
 from pytorch_pfn_extras.training.manager import IgniteExtensionsManager  # NOQA
+from pytorch_pfn_extras.training.metrics import AccuracyMetric  # NOQA
