@@ -1,9 +1,16 @@
-import collections
 import os
 import sys
+import queue
 import time
+from typing import NamedTuple, Optional, Sequence, TextIO, Tuple, TYPE_CHECKING
+
+if TYPE_CHECKING:
+    from pytorch_pfn_extras.training.manager import _BaseExtensionsManager
+
+
 if os.name == 'nt':
     import ctypes
+    from ctypes import windll  # type: ignore [attr-defined]
 
     _STD_OUTPUT_HANDLE = -11
 
@@ -15,24 +22,23 @@ if os.name == 'nt':
                     ('srWindow', ctypes.wintypes.SMALL_RECT),
                     ('dwMaximumWindowSize', _COORD)]
 
-    def set_console_cursor_position(x, y):
+    def set_console_cursor_position(x: int, y: int) -> None:
         """Set relative cursor position from current position to (x,y)"""
 
-        whnd = ctypes.windll.kernel32.GetStdHandle(_STD_OUTPUT_HANDLE)
+        whnd = windll.kernel32.GetStdHandle(_STD_OUTPUT_HANDLE)
         csbi = _CONSOLE_SCREEN_BUFFER_INFO()
-        ctypes.windll.kernel32.GetConsoleScreenBufferInfo(whnd,
-                                                          ctypes.byref(csbi))
+        windll.kernel32.GetConsoleScreenBufferInfo(whnd, ctypes.byref(csbi))
         cur_pos = csbi.dwCursorPosition
         pos = _COORD(cur_pos.X + x, cur_pos.Y + y)
 
         # Workaround the issue that pyreadline overwrites the argtype
-        setpos = ctypes.windll.kernel32.SetConsoleCursorPosition
+        setpos = windll.kernel32.SetConsoleCursorPosition
         argtypes = setpos.argtypes
         setpos.argtypes = None
         setpos(whnd, pos)
         setpos.argtypes = argtypes
 
-    def erase_console(x, y, mode=0):
+    def erase_console(x: int, y: int, mode: int = 0) -> None:
         """Erase screen.
 
         Mode=0: From (x,y) position down to the bottom of the screen.
@@ -40,48 +46,60 @@ if os.name == 'nt':
         Mode=2: Hole screen
         """
 
-        whnd = ctypes.windll.kernel32.GetStdHandle(_STD_OUTPUT_HANDLE)
+        whnd = windll.kernel32.GetStdHandle(_STD_OUTPUT_HANDLE)
         csbi = _CONSOLE_SCREEN_BUFFER_INFO()
-        ctypes.windll.kernel32.GetConsoleScreenBufferInfo(whnd,
-                                                          ctypes.byref(csbi))
+        windll.kernel32.GetConsoleScreenBufferInfo(whnd, ctypes.byref(csbi))
         cur_pos = csbi.dwCursorPosition
         wr = ctypes.c_ulong()
         if mode == 0:
             num = csbi.srWindow.Right * (
                 csbi.srWindow.Bottom - cur_pos.Y) - cur_pos.X
-            ctypes.windll.kernel32.FillConsoleOutputCharacterA(
+            windll.kernel32.FillConsoleOutputCharacterA(
                 whnd, ord(' '), num, cur_pos, ctypes.byref(wr))
         elif mode == 1:
             num = cur_pos.X
-            ctypes.windll.kernel32.FillConsoleOutputCharacterA(
+            windll.kernel32.FillConsoleOutputCharacterA(
                 whnd, ord(' '), num, _COORD(0, cur_pos.Y), ctypes.byref(wr))
         elif mode == 2:
             os.system('cls')
 
 
+class _QueUnit(NamedTuple):
+    iteration: int
+    epoch_detail: float
+    time: float
+
+
 class ProgressBar:
 
-    def __init__(self, out=None):
+    def __init__(self, out: Optional[TextIO] = None) -> None:
         self._out = sys.stdout if out is None else out
-        self._recent_timing = collections.deque([], maxlen=100)
+        self._recent_timing: 'queue.Queue[_QueUnit]' = queue.Queue(maxsize=100)
 
-    def update_speed(self, iteration, epoch_detail):
+    def update_speed(
+            self,
+            iteration: int,
+            epoch_detail: float
+    ) -> Tuple[float, float]:
         now = time.time()
-        self._recent_timing.append((iteration, epoch_detail, now))
-        old_t, old_e, old_sec = self._recent_timing[0]
-        span = now - old_sec
+        self._recent_timing.put(_QueUnit(iteration, epoch_detail, now))
+        old = self._recent_timing.get()
+        span = now - old.time
         if span != 0:
-            speed_t = (iteration - old_t) / span
-            speed_e = (epoch_detail - old_e) / span
+            speed_t = (iteration - old.iteration) / span
+            speed_e = (epoch_detail - old.epoch_detail) / span
         else:
             speed_t = float('inf')
             speed_e = float('inf')
         return speed_t, speed_e
 
-    def get_lines(self):
+    def get_lines(self) -> Sequence[str]:
         raise NotImplementedError
 
-    def update(self, manager=None):
+    def update(
+            self,
+            manager: Optional['_BaseExtensionsManager'] = None
+    ) -> None:
         self.erase_console()
 
         lines = self.get_lines()
@@ -91,23 +109,23 @@ class ProgressBar:
         self.move_cursor_up(len(lines))
         self.flush()
 
-    def close(self):
+    def close(self) -> None:
         self.erase_console()
         self.flush()
 
-    def erase_console(self):
+    def erase_console(self) -> None:
         if os.name == 'nt':
             erase_console(0, 0)
         else:
             self._out.write('\033[J')
 
-    def move_cursor_up(self, n):
+    def move_cursor_up(self, n: int) -> None:
         # move the cursor to the head of the progress bar
         if os.name == 'nt':
             set_console_cursor_position(0, - n)
         else:
             self._out.write('\033[{:d}A'.format(n))
 
-    def flush(self):
+    def flush(self) -> None:
         if hasattr(self._out, 'flush'):
             self._out.flush()
